/*
 * Bundesliga is Copyright (c) 1997-2003 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>
#include <ExgMgr.h>

#include "bl.h"
#include "blrcp.h"
#include "datedb/datedb.h"

/**
 * Database Handling Routinen
 */

DmOpenRef db, datebook;
static SelectedDBInfo dbInfo[maxNumLeagues];
char* Items[maxNumLeagues] = { 0, };
char* theTeams[maxNumTeams];
static Err err = 0;
Int16 numOfItems = 0;

static Int16 DBNameNo(char*) DBFUNC;
static UInt32 dbhash(char* name) DBFUNC;
static Boolean isDB(UInt16 cardNo, LocalID dbID) DBFUNC;
static Err WriteDBData(const void *, UInt32 *, void *) DBFUNC;
static Int16 dbSortFunc(void* a, void* b, Int32 other) DBFUNC;
static void UpgradeDatabaseFromVersionSeven(void) DBFUNC;

/*
 * Create a new Database, install all records.
 */
DmOpenRef CreateDatabase(char* name)
{
  UInt16      index;
  DmOpenRef db;
  LocalID   dbID;
  MemHandle  p;

  if (DBNameNo(name)) {
    DisplayErrorMessage(ID_ErrorMessages+2);
    err = -1;
    return NULL;
  }

  if ( (err = DmCreateDatabase(0, name, 'tpBL', 'Data', false)) )
    return NULL;
  dbID = DmFindDatabase(0, name);
  db = DmOpenDatabase(0, dbID, dmModeReadWrite);

  index = 0;
  p = MemHandleLock(DmNewRecord(db, &index, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, index, true);

  return db;
}

/*
 * Loops over all records and convert them to new format
 */
static void UpgradeDatabaseFromVersionSeven(void)
{
  Int16 i, j;
  Int8 numHandicaps, numTeams;
  char *p, *q;
  matchType m;
  handicapType h;
  UInt32 size;
  UInt16 last;

  for (i=0; i<prefs.NumMatchDays; i++) {
    p = MemHandleLock(DmQueryRecord(db,i+2));
    numTeams = (Int8)*p;
    numHandicaps = (Int8)*(p+1);

    last = dmMaxRecordIndex;
    size = 2 + numTeams*sizeof(matchType) + numHandicaps*sizeof(handicapType);
    q = MemHandleLock(DmNewRecord(db,&last,size));

    DmWrite(q,0,&numTeams,1);
    DmWrite(q,1,&numHandicaps,1);

    for (j=0; j<numTeams; j++) {
      MemMove(&m,p+2+j*sizeof(matchTypeOld),sizeof(matchTypeOld));
      m.penalty = 0;
      DmWrite(q,2+j*sizeof(matchType),&m,sizeof(matchType));
    }

    for (j=0; j<numHandicaps; j++) {
      MemMove(&h,(void*)(p+2+numTeams*sizeof(matchTypeOld)+j*sizeof(handicapType)),
	      sizeof(handicapType));
      DmWrite(q,2+numTeams*sizeof(matchType)+j*sizeof(handicapType),
	      &h,sizeof(handicapType));
    }

    MemPtrUnlock(p);
    MemPtrUnlock(q);

    DmMoveRecord(db,last,i+2);
    DmRemoveRecord(db,i+3);
    DmReleaseRecord(db,i+2,true);
  }

}

/*
 * Read and store contents of Database db "locally".
 */
Err ReadDatabase(DmOpenRef db)
{
  char *p, *q;
  UInt16 i,j,tbSettings = 0;
  
  p = MemHandleLock(DmQueryRecord(db,0));
  MemMove(&prefs, p, sizeof(prefs));
  MemPtrUnlock(p);

  if (prefs.Version < 6)
    return (appErrorClass|tooOld);
  if (prefs.Version < 7) {
    prefs.favTeam = 0;
    prefs.Version = 7;
  }
  if (prefs.Version < 8) {
    UpgradeDatabaseFromVersionSeven();
    prefs.pointsWon     = 3;
    prefs.pointsDraw    = 1;
    prefs.pointsLost    = 0;
    prefs.pointsPenalty = 0;
    prefs.Version = 8;
    if (prefs.options & twoPoints)
      prefs.pointsWon  = 2;
    prefs.tiebreaker[0] = tieBreakerPoints;
    prefs.tiebreaker[1] = tieBreakerGoalsDiff;
    prefs.tiebreaker[2] = tieBreakerGoals;
    prefs.tiebreaker[3] = tieBreakerMatchesWon;
    prefs.tiebreaker[4] = tieBreakerDirectCompare | tieBreakerNotSelected;
    prefs.tiebreaker[5] = tieBreakerPointsDiff    | tieBreakerNotSelected;
    prefs.tiebreaker[6] = tieBreakerPointsPercent | tieBreakerNotSelected;
    prefs.tiebreaker[7] = tieBreakerNone;
    for (i=0; i<5; i++) {
      prefs.colors[i].from  = 0;
      prefs.colors[i].to    = 0;
      prefs.colors[i].color = 255;
    }
  }
  if (prefs.Version < 9) {
    prefs.Version = 9;
    prefs.tableSettings[0] = prefs.tableSettings[5] = tableSettingsMatches;
    prefs.tableSettings[1] = tableSettingsGoals;
    prefs.tableSettings[2] = tableSettingsGoalsAgainst;
    prefs.tableSettings[3] = tableSettingsGoalsDiff;
    prefs.tableSettings[4] = prefs.tableSettings[9] = tableSettingsPts;
    prefs.tableSettings[6] = tableSettingsWon;
    prefs.tableSettings[7] = tableSettingsDraw;
    prefs.tableSettings[8] = tableSettingsLost;
  }

  // add additional new tie breakers and shift !selected flag at bit 7.
  if (prefs.Version < 10) {
    prefs.Version = 10;
    for (i=0; i<7; i++) {
      if (prefs.tiebreaker[i] & 8) {
	prefs.tiebreaker[i] |= tieBreakerNotSelected;
	prefs.tiebreaker[i] ^= 8;
      }
      tbSettings |= (1<<(prefs.tiebreaker[i]&7));
    }
    for (i=0; i<7; i++) {
      if ((prefs.tiebreaker[i]&7) == tieBreakerNone) {
	for (j=0; j<8; j++) {
	  if (! (tbSettings & (1<<j)) ) {
	    prefs.tiebreaker[i] = j | tieBreakerNotSelected;
	    tbSettings |= (1<<j);
	    break;
	  }
	}
      }
    }
  }

  SavePreferences();
  if (TeamResultHandle) {
    MemHandleResize(TeamResultHandle, prefs.NumTeams*sizeof(TeamResultType));
    MemHandleResize(BackupTeamResultHandle, prefs.NumTeams*sizeof(TeamResultType));
  } else {
    TeamResultHandle = MemHandleNew(prefs.NumTeams*sizeof(TeamResultType));
    BackupTeamResultHandle = MemHandleNew(prefs.NumTeams*sizeof(TeamResultType));
  }
  p = q = (char*)MemHandleLock(DmQueryRecord(db,1));
  for (i=0; i<prefs.NumTeams; i++) {
    theTeams[i] = p;
    p += StrLen(p)+1;
  }
  MemPtrUnlock(q);

  return 0;
}

static Boolean isDB(UInt16 cardNo, LocalID dbID)
{
  Int16 i;
  if (numOfItems==0) 
    FindInstalledDatabases();
  for (i=0; i<numOfItems; i++) {
    if ( (dbInfo[i].cardNo == cardNo) &&
         (dbInfo[i].dbID == dbID) )
      return true;
  }
  return false;
}
 

/*
 * Trys to open database dbID.
 */
Err OpenDatabase(UInt16 cardNo, LocalID dbID)
{
  db = 0;

  if (isDB(cardNo,dbID))
    db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
  else
    dbID = 0;

  if (numOfItems == 0) 
    return (appErrorClass|noDatabaseInstalled);

  if (dbID == 0) {
    cardNo = dbInfo[0].cardNo;
    dbID   = dbInfo[0].dbID;
    db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
  }

  if (db) {
    if ( ReadDatabase(db) ) {
      DmCloseDatabase(db);
      db = 0;
      return (appErrorClass|tooOld);
    }
  } else {
    return DmGetLastErr();
  }

  if (!prefs.NumMatchDays)
    prefs.NumMatchDays = 2*(prefs.NumTeams + (prefs.NumTeams%2) -1);

  // validity check; database might be script generated and 
  // contain invalid values.
  if (prefs.matchDay<1 || prefs.matchDay+2 > DmNumRecords(db))
    prefs.matchDay = 1;

  return 0;
}

void SavePreferences(void)
{
  void *p = MemHandleLock(DmResizeRecord(db, 0, sizeof(prefs)));
  DmWrite(p, 0, &prefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);
}

/*
 * Saves program status to open database and updates database
 * "num" attributes.
 */
void SaveStatus(Int16 num)
{
  Int16 i;
  UInt16 size;
  
  prefs.matchDay = matchDay;

  if (!db)
    return;

  SavePreferences();

  dbInfo[num].dbAttr |= dmHdrAttrBackup;
  DmSetDatabaseInfo (dbInfo[num].cardNo, dbInfo[num].dbID, NULL, 
		     &dbInfo[num].dbAttr, 
		     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  DmCloseDatabase(db);

  if (appPrefs.options & dontSaveEM)
    appPrefs.options &= ~editMode;

  /* remember sort order the user chose for the leagues */
  for (i = 0; i<maxNumLeagues; i++) {
    if (i<numOfItems) {
      appPrefs.dbHash[i] = dbhash(Items[i]);
    } else {
      appPrefs.dbHash[i] = 0;
    }
  }

  appPrefs.Version = 3;
  size = sizeof(appPrefs);
  PrefSetAppPreferences('tpBL', 0, appPrefs.Version, &appPrefs, size, false);
}

void DeleteRecord(DmOpenRef db, UInt16 index)
{
  DmRemoveRecord(db,index);
}

void InsertNewRecord(DmOpenRef db, UInt16 *index, UInt16 numMatches)
{
  char *p;
  matchType m;
  UInt16 j = numMatches<<8;

  p = MemHandleLock(DmNewRecord(db, index, sizeof(matchType)*numMatches+2));
  DmWrite(p,0,&j,2);
  for (j=0; j<numMatches; j++) {
    m.t1 = j*2+1; m.t2=j*2+2; m.r1=255; m.r2=255; m.Date=0; m.penalty=0;
    DmWrite(p,2+sizeof(matchType)*j,&m,sizeof(matchType));
  }
  MemPtrUnlock(p);
  DmReleaseRecord(db, *index, true);
}

/*
 * write "sane" values to freshly created database db
 */
void InitNewDatabase(DmOpenRef db, BlDatabasePreferencesType* newPrefs)
{
  char *p, buf[8];
  UInt16 i, index;

  index = 0;
  p = MemHandleLock(DmNewRecord(db, &index, sizeof(prefs)));
  DmWrite(p, 0, newPrefs, sizeof(prefs));
  MemPtrUnlock(p);
  DmReleaseRecord(db, 0, true);

  StrCopy(buf,"Team A");
  index = 1;
  p = MemHandleLock(DmNewRecord(db, &index, 7*newPrefs->NumTeams));
  for (i=0; i<newPrefs->NumTeams; i++) {
    buf[5] = 'A'+i;
    DmWrite(p,i*7,buf,7);
  }
  MemPtrUnlock(p);
  DmReleaseRecord(db, 1, true);

  for (i=0; i<newPrefs->NumMatchDays; i++) {
    index = i+2;
    InsertNewRecord(db, &index, newPrefs->NumTeams>>1);
  }

  numOfItems = 0;
}

/*
 * looks for installed databases and stores result into global dbInfo[].
 */
Int16 FindInstalledDatabases(void)
{
  void*   RecPointer;
  DmOpenRef db;
  BlDatabasePreferencesType newPrefs;
  Boolean newSearch = true;
  DmSearchStateType stateInfo;
  Int16 i = 0;

  while (i<maxNumLeagues && DmGetNextDatabaseByTypeCreator(newSearch, &stateInfo, 
	 'Data', 'tpBL', false, &dbInfo[i].cardNo, &dbInfo[i].dbID) == 0) {
    newSearch = false;

    db = DmOpenDatabase(dbInfo[i].cardNo, dbInfo[i].dbID, dmModeReadOnly);
    RecPointer = MemHandleLock(DmQueryRecord(db, 0));
    MemMove(&newPrefs, RecPointer, sizeof(newPrefs));
    MemPtrUnlock(RecPointer);
    DmCloseDatabase(db);
  
    if (newPrefs.Version < 6) {
      continue;
    }
    i++;
  }
  numOfItems = i;
  return numOfItems;
}  

/*
 * returns number of database with name "name"
 */
static Int16 DBNameNo(char* name)
{
  Int16 i;
  for (i=0; Items[i]; i++) {
    if (!StrCompare(name,Items[i]))
      return i+1;
  }
  return 0;
}

/* Sorting function for use with SysInsertionSort */
static Int16 dbSortFunc(void* a, void* b, Int32 other)
{
  UInt32 keyA = dbhash(((SelectedDBInfo*)a)->dbName);
  UInt32 keyB = dbhash(((SelectedDBInfo*)b)->dbName);
  Int16 i;

  for (i=0; i<maxNumLeagues; i++) {
    if (keyA == appPrefs.dbHash[i])
      return -1;
    if (keyB == appPrefs.dbHash[i])
      return 1;
  }
  return 0;
}

void ChooseDB(void)
{
  Int16 i;
  ListPtr p;

  if (numOfItems == 0)
    FindInstalledDatabases();

  for (i=0; i<numOfItems; i++) {
    DmDatabaseInfo(dbInfo[i].cardNo, 
		   dbInfo[i].dbID, 
		   dbInfo[i].dbName, 
		   NULL, NULL, NULL, NULL, NULL, 
		   NULL, NULL, NULL, NULL, NULL);
  }

  /* preserve choosen sort order */
  SysInsertionSort(dbInfo,numOfItems,sizeof(SelectedDBInfo),dbSortFunc,0);

  for (i=0; i<numOfItems; i++) {
    Items[i] = dbInfo[i].dbName;

    /* this must appear _after_ sorting the items. Having it here
     * doesn't hurt, because appPrefs.dbID is consistently updated
     */
    if ( (appPrefs.dbID == dbInfo[i].dbID) && 
	 (appPrefs.cardNo == dbInfo[i].cardNo) ) 
	 DBNo = i;
  }

  p = FrmGetListPtr(ID_FrmBl, ID_FrmPopup+1);
  LstSetListChoices(p, Items, numOfItems);
  LstSetHeight(p, numOfItems>10?10:numOfItems);
}

void SelectedDB(Int16 i, SelectedDBInfo *info)
{
  MemMove(info, &dbInfo[i], sizeof(SelectedDBInfo));
  appPrefs.dbID = dbInfo[i].dbID;
  appPrefs.cardNo = dbInfo[i].cardNo;
}

Err DeleteDatabase(Int16 i)
{
  Err err;
  Int16 j;
  err = DmCloseDatabase(db);
  if ((err = DmDeleteDatabase(dbInfo[i].cardNo, dbInfo[i].dbID))) {
    DisplayErrorMessage(ID_ErrorMessages+3);
    return err;
  }
  for (j=i; j<numOfItems-1; j++) {
    MemMove(&dbInfo[j], &dbInfo[j+1], sizeof(dbInfo[j]));
  }
  numOfItems--;
  if (numOfItems == 0) {
    db = 0;
    return (appErrorClass|noDatabaseInstalled);
  }
  ChooseDB();
  return 0;
}

Boolean RenameDatabase(Int16 i, char* name)
{
  if (DmSetDatabaseInfo (dbInfo[i].cardNo, dbInfo[i].dbID, 
			 name, NULL, NULL, NULL, NULL, NULL, 
			 NULL, NULL, NULL, NULL, NULL) == 0) {
    StrCopy(dbInfo[i].dbName,name);
    return true;
  }
  return false;
}

/* 
 * Hash function that generates keys from database names. The keys rather
 * than the database names are stored in the application preferences.
 */
static UInt32 dbhash(char* name)
{
  UInt32 key = 0;
  Int8 c;

  while ((c = *name++)) {
    key += c;
    key ^= (key << 24) | (key << 16) | (key << 8);
  }

  return key;
}

/*
 * exchange two dbInfo entries
 */
void swapDbInfo(Int16 i, Int16 j)
{
  SelectedDBInfo tmp = dbInfo[i];
  dbInfo[i] = dbInfo[j];
  dbInfo[j] = tmp;
}

static Int16 SortAlphabetical(void* a, void* b, Int32 arg)
{
  return StrCaselessCompare((const char*)((SelectedDBInfo*)a)->dbName,
			    (const char*)((SelectedDBInfo*)b)->dbName);
}


void SortDbInfo(void)
{
  Int16 i;
  LocalID dbID;

  dbID = dbInfo[DBNo].dbID;
  SysQSort(dbInfo,numOfItems,sizeof(SelectedDBInfo),SortAlphabetical,0);
  for (i=0; i<numOfItems; i++) {
    Items[i] = dbInfo[i].dbName;
    if (dbInfo[i].dbID == dbID)
      DBNo = i;
  }
}

/*
 * Beaming support; we transfer a complete database and let 
 * the launcher application do the dirty work :-)
 */
static Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP)
{
  Err err;
  *sizeP = ExgSend((ExgSocketType*)userDataP,(void*)dataP,*sizeP,&err);
  return err;
}

void BeamDatabase(Int16 no)
{
  ExgSocketType s;
  char name[36];
  Err err;

  if (!FeatureSet30) {
    DisplayInformationMessage(ID_InformationMessages+1);
    return;
  }
  
  StrCopy(name, dbInfo[no].dbName);
  StrCat(name, ".pdb");

  MemSet(&s, sizeof(s), 0);
  s.description = name;
  s.name = name;

  if (! (err = ExgPut(&s)) ) {
    err = ExgDBWrite(WriteDBData,&s,NULL,dbInfo[no].dbID,dbInfo[no].cardNo);
  }
  ExgDisconnect(&s, err);

}

/*
 * Datebook interface functions
 */
Boolean OpenDatebook(void)
{
  LocalID id = 0;
  DmSearchStateType state; 
  UInt16 card;
  Err theErr;

  /*
   * Try to open the datebook database
   */
  datebook = 0;
  theErr = DmGetNextDatabaseByTypeCreator(true, &state, 'DATA', 
                                          'date', true, &card, &id); 
  if ( (theErr != dmErrCantFind) ) {
    datebook = DmOpenDatabase(card, id, dmModeReadWrite);
    return true;
  }
  return false;
}

/*
 * Arguments: 
 * match: all relevant match data.
 */
void MakeDatebookEntry(matchType* match)
{
  ApptDBRecordType dbrec;
  ApptDateTimeType dbwhen;
  AlarmInfoType dbalarm;

  DateTimeType date_time;
  UInt16 index;
  Char buf[64];

  /*
   * no date or invalid teams => no entry
   */
  if (match->Date == 0 || match->t1 == 0 || match->t2 == 0) 
    return;

  /*
   * initializing Datebook structures
   */
  MemSet(&dbrec, sizeof(ApptDBRecordType), 0);
  MemSet(&dbwhen, sizeof(ApptDateTimeType), 0);
  MemSet(&dbalarm, sizeof(AlarmInfoType), 0);

  dbrec.when = &dbwhen;
  if (appPrefs.options & setAlarm)
    dbrec.alarm = &dbalarm;
  else
    dbrec.alarm = NULL;
  dbrec.repeat = NULL;
  dbrec.note = NULL;
  dbrec.description = NULL;

  /*
   * building the match description
   */
  StrCopy(buf,theTeams[match->t1-1]);
  StrCat(buf,"-");
  StrCat(buf,theTeams[match->t2-1]);
  dbrec.description = buf;
  
  /*
   * setting the match time
   */
  DateSecondsToDate(match->Date,&dbwhen.date);
  TimSecondsToDateTime(match->Date, &date_time);
  // if (match->Date % 60)
  //   TimAdjust(&date_time, (60 - (match->Date % 60)));
  dbwhen.startTime.hours = date_time.hour;
  dbwhen.startTime.minutes = date_time.minute;
  if (dbwhen.startTime.hours < 22) {
    dbwhen.endTime.hours = dbwhen.startTime.hours+2;
    dbwhen.endTime.minutes = dbwhen.startTime.minutes;
  } else {
    dbwhen.endTime.hours = 23;
    dbwhen.endTime.minutes = 59;
  }

  dbalarm.advanceUnit = aauMinutes;
  dbalarm.advance = appPrefs.minutesAlarm;

  ApptNewRecord(datebook, &dbrec, &index);
}

