/*
 * Bundesliga is Copyright (c) 1997-2003 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>

#include "bl.h"
#include "blrcp.h"

typedef struct {
    UInt16                  formID;
    FormEventHandlerType*   handler;
} EventHandler;

static EventHandler ListOfEventHandlers[] = {
  { ID_FrmBl, MatchdayFormHandleEvent },
  { ID_FrmTable, TableFormHandleEvent },
  { ID_FrmTableSettings, TableSettingsFormHandleEvent },
  { ID_FrmGoto, GotoFormHandleEvent },
  { ID_FrmAbout, AboutFormHandleEvent },
  { ID_FrmStatistics, StatisticsFormHandleEvent },
  { ID_FrmEdit, EditFormHandleEvent },
  { ID_FrmTeams,TeamsFormHandleEvent  },
  { ID_FrmSettings, SettingsFormHandleEvent },
  { ID_FrmDatabaseSettings, DatabaseSettingsFormHandleEvent },
  { ID_FrmDatabaseColors, DatabaseColorsFormHandleEvent },
  { ID_FrmEnterTeam, EnterTeamFormHandleEvent },
  { ID_FrmRenameLeague, RenameLeagueFormHandleEvent },
  { ID_FrmNewDB, NewDBFormHandleEvent },
  { ID_FrmHandicap, HandicapFormHandleEvent },
  { ID_FrmHandicapEdit, HandicapEditFormHandleEvent },
  { ID_FrmLeague, LeagueFormHandleEvent },
  { ID_FrmShowFixture, ShowFixtureFormHandleEvent },
  { 0, 0 },
};

#ifndef SUPPORT_DIA
static void DiaDummy(void) __attribute__ ((__section__("diafns")));
static void DiaDummy(void) {}
#endif

/**
 * Main Eventhandler Routinen
 */

UInt32 romversion;

static Boolean ApplicationHandleEvent(EventPtr e)
{
  EventHandler *list;

  if (e->eType==frmLoadEvent) {
    UInt16 frmId = e->data.frmLoad.formID;
    FormPtr frm = FrmInitForm(frmId);

    FrmSetActiveForm(frm);

    for (list=ListOfEventHandlers; list->formID != 0; list++ ) {
      if (frmId == list->formID) {
	SetResizePolicy(frmId);
	FrmSetEventHandler(frm, list->handler);
	return true;
      }
    }
  }
  return false;
}

static Int16 StartApplication(void)
{
  UInt16 size;
  Int16 i;
  Err err;
  
  size = sizeof(appPrefs);
  if ( !PrefGetAppPreferences('tpBL',0,&appPrefs,&size,false) ) {
    appPrefs.dbID         = 0;
    appPrefs.cardNo       = 0;
    appPrefs.Version      = 3;
    appPrefs.options      = 0;
    appPrefs.minutesAlarm = 30;
    for (i=0; i<5; i++) {
      appPrefs.color[i] = 255;
    }
  }
  
  if ( (err = OpenDatabase(appPrefs.cardNo, appPrefs.dbID)) ) {
    if ( err == (appErrorClass|tooOld) )
      DisplayErrorMessage(ID_ErrorMessages+1);
    if ( err == (appErrorClass|noDatabaseInstalled) )
      DisplayErrorMessage(ID_ErrorMessages+5);
    return err;
  }

  if (appPrefs.Version < 2)
    appPrefs.minutesAlarm = 30;

  if (appPrefs.Version < 3)
    for (i=0; i<5; i++) {
      appPrefs.color[i] = 255;
    }

  matchDay = prefs.matchDay;
  matchSymbol = ( appPrefs.options & 24) >> 3;

  // initialize DIA API support.
  InitializeResizeSupport(ID_indexDIADataID);

  return 0;
}

static void EventLoop(void)
{
  EventType e;
  short err;

  do {
    EvtGetEvent(&e, evtWaitForever);

    if (! SysHandleEvent (&e))
      if (! MenuHandleEvent (NULL, &e, &err))
        if (! ApplicationHandleEvent (&e))
          FrmDispatchEvent (&e);
  } while (e.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
  Err err;
  
  if (sysAppLaunchCmdNormalLaunch == cmd) {

    err = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romversion);
    if (FeatureSet20) {

      if (StartApplication())
	return 0;

      if (FeatureSet35) {
        // find out what color depths the device supports
	UInt32 depth = 0;
        err = WinScreenMode (winScreenModeGetSupportedDepths,
                             NULL, NULL, &depth, NULL );
      
        // older 3.x ROM's seem to only support 4 color greyscale
        if (depth<4) {
          depth = 2;
        } else if (depth<16) {
          depth = 4;
        }
        // change color depth
        err = WinScreenMode( winScreenModeSet, 
                             NULL, NULL, &depth, NULL );
      }

      FrmGotoForm(ID_FrmBl);
      
      EventLoop();
      
      FrmCloseAllForms();
      SaveStatus(DBNo);

      if (FeatureSet35) {
        WinScreenMode( winScreenModeSetToDefaults, 
                       NULL, NULL, NULL, NULL );
      }

      if (TeamResultHandle) {
	MemHandleFree(TeamResultHandle);
	MemHandleFree(BackupTeamResultHandle);
      }

      if (progDatePtr)
        MemPtrFree(progDatePtr);

      TerminateResizeSupport();

    } else {

      DisplayInformationMessage(ID_InformationMessages);
      AppLaunchWithCommand(sysFileCDefaultApp,
			   sysAppLaunchCmdNormalLaunch, NULL);
    }

  } else if (sysAppLaunchCmdNotify == cmd) {

    SysNotifyParamType *np = (SysNotifyParamType *)cmdPBP;
    HandleResizeNotification(np->notifyType);

  }


  return 0;
}
