/*
 * Bundesliga is Copyright (c) 1997-2003 by Thomas Pundt
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice and this paragraph and the
 * following two paragraphs appear in all copies.
 *
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
 * BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include <PalmOS.h>

#include "bl.h"
#include "blrcp.h"

/*
 * helper function to acquire a handle if necessary and set contents
 * of a field.
 */
void SetField(UInt16 ID, Int16 size, char* val)
{
  FieldType* fld;
  MemHandle handle;
  char*  name;

  fld = Id2Ptr(ID);
  handle = FldGetTextHandle(fld);
  FldSetTextHandle(fld, NULL);
  if (!handle) {
    handle = MemHandleNew(size*sizeof(char));
  } else {
    MemHandleResize(handle,size*sizeof(char));
  }
  name = MemHandleLock(handle);
  StrCopy(name, val);
  MemHandleUnlock(handle);
  FldSetTextHandle(fld, handle);
}

/**
 * UI helper function(s)
 */

void* Id2Ptr(UInt16 id)
{ 
  FormPtr frm = FrmGetActiveForm();
  return FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,id));
}

void FrmMoveObject(Int16 objId, Int16 xd, Int16 yd) 
{
  FormType *frm = FrmGetActiveForm();
  UInt16 x,y;
  UInt16 objIndex = FrmGetObjectIndex(frm ,objId);
  FrmGetObjectPosition(frm, objIndex, &x, &y);
  FrmSetObjectPosition(frm, objIndex, x+xd, y+yd);
}

void FrmResizeObject(Int16 objId, Int16 width, Int16 height) 
{
  FormType *frm = FrmGetActiveForm();
  RectangleType bounds;
  UInt16 objIndex = FrmGetObjectIndex(frm ,objId);
  FrmGetObjectBounds(frm, objIndex, &bounds);
  bounds.extent.x = width;
  bounds.extent.y = height;
  FrmSetObjectBounds(frm, objIndex, &bounds);
}

ListPtr FrmGetListPtr(UInt16 frm, UInt16 id)
{
  FormPtr frmPtr = FrmGetFormPtr(frm);
  return FrmGetObjectPtr(frmPtr, FrmGetObjectIndex(frmPtr, id));
}

void ButtonSetLabel(Int16 id, char *l)
{
  ControlType *ctlP = Id2Ptr(id);
  char *label;
  if (l) {
    label = (char*)CtlGetLabel(ctlP);
    CtlEraseControl(ctlP);
    StrCopy(label,l);
    CtlDrawControl(ctlP);
  }
}

char* GetString(UInt16 id)
{
  char *p;
  MemHandle h = DmGetResource('tSTR', id);
  if (!h) return NULL;
  p = MemHandleLock(h);
  MemHandleUnlock(h);
  return p;
}

void DisplayErrorMessage(UInt16 id)
{
  char *err = GetString(id);
  FrmCustomAlert(ID_ErrorMessagesForm,err,NULL,NULL);
}

void DisplayInformationMessage(UInt16 id)
{
  char *err = GetString(id);
  FrmCustomAlert(ID_InformationMessagesForm,err,NULL,NULL);
}

void DisplayErrorMessageAndStop(UInt16 id)
{
  EventType e;
  DisplayErrorMessage(id);
  e.eType = appStopEvent;
  EvtAddEventToQueue(&e);
}

UInt16 DisplayConfirmationMessage(UInt16 id)
{
  char *msg = GetString(id);
  return FrmCustomAlert(ID_ConfirmationMessagesForm,msg,NULL,NULL);
}

/*
 * Graphical buttons are only available since OS 3.5;
 * we emulate them using WinDrawBitmap().
 * Parameters: res        Resource ID of control;
 *             bitmapid   Id of bitmap
 */
void DrawButton(Int16 res, Int16 bitmapid)
{
  RectangleType rect;
  FormPtr frm;
  void *hBmp;

  frm = FrmGetActiveForm();
  FrmGetObjectBounds(frm,FrmGetObjectIndex(frm,res),&rect);
  CtlShowControl(Id2Ptr(res));
  if ( (hBmp = DmGetResource((UInt32)'Tbmp',bitmapid)) ) {
    WinDrawBitmap(MemHandleLock(hBmp),
                  rect.topLeft.x,rect.topLeft.y);
    MemHandleUnlock(hBmp);
    DmReleaseResource(hBmp);
  }
}

/*
 * Reposition/recalculate scrollbar for a field.
 */
void FldRecalculateScrollbar(FieldPtr fld, ScrollBarPtr scl)
{
  UInt16 scrollPos, textHeight, fieldHeight;
  Int16 maxScrollPos;
  
  // Get the current scroll position of the field, the total number of
  // lines in the field, and the number of lines visible in the field
  FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);
  
  // Calculate the maximum scroll position:
  if(textHeight > fieldHeight)
    maxScrollPos = textHeight - fieldHeight + 1;
  else
    maxScrollPos = scrollPos;
  
  // Set the scroll bar's position:
  // Use the page size fieldHeight-1; this sets the size of the scroll car,
  // and gives an overlap of 1 line when page scrolling.
  SclSetScrollBar(scl, scrollPos, 0, maxScrollPos, fieldHeight-1);
}

/*
 * Do all the computing and scrolling.
 */
void FldDoScrolling(FieldPtr fld, ScrollBarPtr scl)
{
  Int16 value, min, max, pageSize;
  UInt16 scrollPos, textHeight, fieldHeight;
        
  // Get the current position of the scroll bar:
  SclGetScrollBar(scl, &value, &min, &max, &pageSize);
  
  // Get the current scroll position of the field:
  FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);
  
  // Scroll the field according to the scroll bar's position:
  if(scrollPos > value)
    FldScrollField(fld, scrollPos-value, winUp);
  else if(scrollPos < value)
    FldScrollField(fld, value-scrollPos, winDown);
}

/*
 * Init function for HelpForm
 */
void InitFrmHelp(UInt16 ID_tSTR)
{
  FormPtr frm = FrmGetActiveForm();
  FieldPtr fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ID_FrmHelp));
  static char* msg;
  msg = GetString(ID_tSTR);
  FldSetTextPtr(fld, msg);
  FldRecalculateField(fld, true);
  FldRecalculateScrollbar(fld, Id2Ptr(ID_FrmHelp+1));
}

/*
 * Event handler for HelpForm
 */
Boolean FrmHelpHandleEvent (EventPtr e)
{
  Boolean handled = false;
  
  if ( (handled = ResizeHandleEvent(e)) ) {
    return true;
  }

  switch (e->eType) {
  case sclRepeatEvent:
    FldDoScrolling(Id2Ptr(ID_FrmHelp), Id2Ptr(ID_FrmHelp+1));
    break;
  
  default:
    ;
  }
  return handled;
}

/*
 * Show a modal dialog; set event handler and use a init function if given.
 */
void MyDoDialog(UInt16 frmID, UInt16 strID, char *title,
                FormEventHandlerType *evtHandler,
                FormInitFunctionType *initForm)
{
  FormPtr savedForm = FrmGetActiveForm(),
          frm       = FrmInitForm(frmID);
  FrmSetActiveForm(frm);
  if (evtHandler) {
    SetResizePolicy(frmID);
    FrmSetEventHandler(frm, evtHandler);
  }
  if (initForm) 
    initForm(strID);
  if (title)
    FrmSetTitle(frm,title);
  ResizeRefreshCurrentForm();
  FrmDrawForm(frm);
  FrmDoDialog(frm);
  FrmDeleteForm(frm);
  FrmSetActiveForm(savedForm);
}
